import 'package:after_layout/after_layout.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:osam/osam.dart';

class OsamNavigator extends StatefulWidget {
  final Map<String, Widget> screens;
  final PresenterWithNavigation presenter;
  final bool rootNavigator;
  final String initialRoute;
  final bool isMaterial;

  const OsamNavigator(
      {Key key,
      this.screens,
      this.presenter,
      this.rootNavigator = false,
      this.initialRoute,
      this.isMaterial = false})
      : super(key: key);

  @override
  _OsamNavigatorState createState() => _OsamNavigatorState();
}

class _OsamNavigatorState extends State<OsamNavigator> with AfterLayoutMixin {
  final List<String> _routes = [];

  Map<String, Widget> get screens => widget.screens;

  PresenterWithNavigation get presenter => widget.presenter;

  @override
  Widget build(BuildContext context) => widget.screens[widget.initialRoute];

  void _silentPop(dynamic route) {
    if (presenter.routeStream.initialData.length > 1) {
      _routes.removeLast();
      presenter.pop();
    }
  }

  PageRoute pageRoute({WidgetBuilder builder, RouteSettings settings}) => widget.isMaterial
      ? MaterialPageRoute(builder: builder, settings: settings)
      : CupertinoPageRoute(builder: builder, settings: settings);

  @override
  void afterFirstLayout(BuildContext context) {
    final navigator = Navigator.of(context, rootNavigator: widget.rootNavigator);
    final initialRoutes = presenter.routeStream.initialData;
    _routes.addAll(initialRoutes);
    if (initialRoutes.length > 1) {
      final firstRoute = initialRoutes.first;
      navigator.pushReplacement(pageRoute(
          builder: ((ctx) => screens[firstRoute]), settings: RouteSettings(isInitialRoute: true)));
      final others = initialRoutes.sublist(1, initialRoutes.length);
      others.forEach((route) {
        navigator.push(pageRoute(builder: ((ctx) => screens[route]))..popped.then(_silentPop));
      });
    }
    presenter.routeStream.stream.listen((routes) {
      final routesDifferentNumber = routes.length - _routes.length;
      if (routesDifferentNumber == 0 && routes.last != _routes.last) {
        navigator.pushReplacement(pageRoute(builder: ((ctx) {
          return screens[routes.last];
        }))
          ..popped.then(_silentPop));
      } else if (routesDifferentNumber > 0) {
        final pushRoutes = routes.sublist(_routes.length, routes.length);
        pushRoutes.forEach((route) {
          navigator.push(pageRoute(builder: ((ctx) => screens[route]))..popped.then(_silentPop));
        });
      } else if (routesDifferentNumber < 0) {
        _routes.first != routes.first
            ? navigator.pushAndRemoveUntil(pageRoute(builder: ((ctx) {
                return screens[routes.first];
              })), (route) => false)
            : _routes.sublist(routes.length, _routes.length).forEach((route) async {
                navigator.pop();
              });
      }
      _routes.clear();
      _routes.addAll(routes);
    });
  }
}

class OsamTabNavigator extends StatelessWidget {
  final Map<String, OsamNavigator> navigators;
  final Map<String, Widget> tabs;
  final PresenterWithTabNavigation presenter;
  final BoxDecoration tabDecoration;

  const OsamTabNavigator({Key key, this.navigators, this.presenter, this.tabDecoration, this.tabs})
      : super(key: key);

  @override
  Widget build(BuildContext context) {
    return OsamBuilder<String>(
      stream: presenter.tabStream,
      builder: (ctx, snapshot) {
        return Stack(
          children: <Widget>[
            IndexedStack(
              index: navigators.keys.toList().indexOf(snapshot.data),
              children: navigators.keys.map((tab) {
                return _OsamNavigatorPage(
                  key: ValueKey(tab.toString() + '_onp'),
                  navKey: GlobalKey<NavigatorState>(),
                  child: navigators[tab],
                );
              }).toList(),
            ),
            Positioned(
              bottom: 0,
              child: SafeArea(
                bottom: true,
                child: Container(
                  decoration: tabDecoration,
                  padding: const EdgeInsets.symmetric(vertical: 5),
                  constraints: BoxConstraints(
                    maxWidth: MediaQuery.of(context).size.width,
                    minHeight: 48,
                  ),
                  child: Row(
                    children: navigators.keys
                        .map((tab) => Expanded(
                              child: GestureDetector(
                                child: Container(
                                  child: tabs[tab],
                                  alignment: Alignment.center,
                                  color: Colors.transparent,
                                ),
                                onTap: () {
                                  if (snapshot.data != tab) {
                                    presenter.changeTab(tab);
                                  } else {
                                    final currentPresenter = navigators[snapshot.data].presenter;
                                    final routes = currentPresenter.routeStream.initialData;
                                    if (routes.length > 1) {
                                      currentPresenter.pushAndRemoveUntil(routes.first);
                                    }
                                  }
                                },
                              ),
                            ))
                        .toList(),
                  ),
                ),
              ),
            )
          ],
        );
      },
    );
  }
}

class _OsamNavigatorPage extends StatelessWidget {
  final GlobalKey<NavigatorState> navKey;
  final Widget child;

  const _OsamNavigatorPage({Key key, this.navKey, this.child}) : super(key: key);

  @override
  Widget build(BuildContext context) {
    return Navigator(
      key: navKey,
      onGenerateRoute: (RouteSettings settings) => MaterialPageRoute(
        settings: RouteSettings(isInitialRoute: true),
        builder: (BuildContext context) => child,
      ),
    );
  }
}
